# Thanks to
# https://stackoverflow.com/a/43573798/
# https://stackoverflow.com/a/45168493/
# https://stackoverflow.com/a/48671982/

import sys
import inspect
from types import CodeType
from threading import Lock
from typing import Callable, Optional
from importlib import invalidate_caches
from importlib.util import find_spec
from importlib.machinery import (
    BYTECODE_SUFFIXES,
    SOURCE_SUFFIXES,
    FileFinder,
    PathFinder as _PathFinder,
    SourceFileLoader,
)

from .config import FILE_EXTENSIONS
from .preprocessor import PyPreprocessor, preprocess_file
from .utils import import_module_copy, create_sys_clone
from .monkeypatch import (
    apply_monkeypatch,
    dependencies,
)


# very hacky way to replace `sys` in the PathFinder with our own module
# this is done because we need separate `path_hooks` and `path_importer_cache`
original_pathfinder_module = inspect.getmodule(_PathFinder)
pathfinder_module = import_module_copy(original_pathfinder_module.__name__)
PathFinder = pathfinder_module.PathFinder
vars(pathfinder_module).update(vars(original_pathfinder_module))
pathfinder_module.sys = create_sys_clone()


class PPyPathFinder(PathFinder):
    """
    An overridden PathFinder which will hunt for ppy files in sys.path
    """

    prefer_python = False

    @classmethod
    def find_spec(
        cls, fullname: str, path: Optional[list] = None, target=None
    ):
        if cls.prefer_python:
            index = sys.meta_path.index(cls)
            del sys.meta_path[index]
            try:
                spec = find_spec(fullname)
            finally:
                sys.meta_path.insert(index, cls)
            if spec:
                return spec

        spec = super().find_spec(fullname, path, target)
        if spec is not None and spec.loader is not None:
            return spec
        return None


class PPyLoader(SourceFileLoader):
    save_files = False
    _skip_next_get_data = False
    _get_code_lock = Lock()

    def get_data(self, filename: str) -> Optional[bytes]:
        if self._skip_next_get_data:
            self.__class__._skip_next_get_data = False
            return None

        if filename.endswith(tuple(BYTECODE_SUFFIXES)):
            with open(filename, "rb") as f:
                return f.read()

        data, deps = preprocess_file(self.path, self.save_files)
        dependencies[self.path] = deps

        return data.encode()

    def source_to_code(self, data: bytes, path: str) -> CodeType:
        code = super().source_to_code(data, path)
        if self.path in dependencies:
            dependencies[code] = dependencies.pop(self.path)
        return code

    def get_code(self, fullname: str) -> CodeType:
        with self._get_code_lock:
            return super().get_code(fullname)


LOADER_DETAILS = PPyLoader, FILE_EXTENSIONS


def _install() -> Callable[..., None]:
    done = False

    def install(
        *,
        save_files: bool,
        prefer_python: bool,
        preprocess_unknown_sources: bool,
    ):
        nonlocal done

        # (re)setting global configuration
        PPyLoader.save_files = save_files
        PPyPathFinder.prefer_python = prefer_python
        PyPreprocessor.default_disabled = not preprocess_unknown_sources

        if done:
            return

        # insert the path finder
        sys.meta_path.insert(0, PPyPathFinder)
        pathfinder_module.sys.path_hooks.append(
            FileFinder.path_hook(LOADER_DETAILS)
        )
        # register our extension
        SOURCE_SUFFIXES.extend(FILE_EXTENSIONS)
        # clear any loaders that might already be in use by the FileFinder
        sys.path_importer_cache.clear()
        invalidate_caches()
        # patch standard library
        apply_monkeypatch()

        done = True

    return install


install = _install()



import os
import sys
import time
from io import StringIO
from unittest.mock import patch
from subprocess import STDOUT, CalledProcessError, check_output

import pytest

sys.path.insert(0, os.path.dirname(os.path.dirname(__file__)))

from pwcp import main  # noqa: E402
from pwcp.utils import is_package  # noqa: E402


sys.dont_write_bytecode = True
os.environ["PYTHONDONTWRITEBYTECODE"] = "1"


def test_regular_file():
    with patch("sys.stdout", new=StringIO()):
        main(["tests/hello.py"])
        assert sys.stdout.getvalue() == "Hello world! (python)\n"


def test_ppy_file():
    with patch("sys.stdout", new=StringIO()):
        main(["tests/hello.ppy"])
        assert sys.stdout.getvalue() == "Hello world!\nNone world!\n"


def test_run_module():
    with patch("sys.stdout", new=StringIO()):
        main(["-m", "tests.a_module", "1", "2", "3"])
        assert sys.stdout.getvalue() == "tests.a_module.b = 6\n"


def test_comments():
    main(["tests/comments.ppy"])


def test_imports():
    main(["tests/test_modules.ppy"])
    assert sys.modules["hello"].__file__ == os.path.join(
        os.path.abspath("tests"), "hello.ppy"
    )
    del sys.modules["hello"]


def test_py_import():
    main(["--prefer-py", "tests/test_modules.ppy"])
    assert sys.modules["hello"].__file__ == os.path.join(
        os.path.abspath("tests"), "hello.py"
    )
    del sys.modules["hello"]


def test_syntax_error():
    with pytest.raises(CalledProcessError) as ctx:
        check_output(
            [sys.executable, "-m", "pwcp", "tests/syntax_error.ppy"],
            stderr=STDOUT,
        )
    assert ctx.value.output.splitlines()[-3].strip() == b'print("hello")!'


def test_type_error():
    with pytest.raises(CalledProcessError) as ctx:
        check_output(
            [sys.executable, "-m", "pwcp", "tests/type_error.ppy"],
            stderr=STDOUT,
        )
    assert ctx.value.output.splitlines()[2].strip() == b"print('1' + 1)"


def test_error_directive():
    with pytest.raises(CalledProcessError) as ctx:
        check_output(
            [sys.executable, "-m", "pwcp", "tests/error_directive.ppy"],
            stderr=STDOUT,
        )
    assert (
        ctx.value.output.splitlines()[-1].strip()
        == b"pwcp.errors.PreprocessorError:"
        b" preprocessor exit code is not zero: 1"
    )


def test_interactive():
    s = "\nand this is a triple-quoted string\n"
    code = f"""
#define x 1
/* this
is
a comment */
'''{s}'''
def f():
    '''check whether we can declare
    a function with several lines'''
#if x
    print(x)
#endif

f()
    """.strip()
    with patch("sys.stdin", new=StringIO(code)), patch(
        "sys.stdout", new=StringIO()
    ):
        main(["--preprocess-unknown-sources", "-m", "code"])
        assert (
            sys.stdout.getvalue()
            == sys.ps1 * 2
            + sys.ps2 * 2
            + sys.ps1
            + sys.ps2 * 2
            + repr(s)
            + "\n"
            + sys.ps1
            + sys.ps2 * 6
            + sys.ps1
            + "1\n"
            + sys.ps1
        )

    code = """
#ifndef X
#define X
print(1)
#endif
    """.strip()
    with patch("sys.stdin", new=StringIO(code)), patch(
        "sys.stdout", new=StringIO()
    ):
        main(["--preprocess-unknown-sources", "-m", "code"])
        assert sys.stdout.getvalue() == sys.ps1 + sys.ps2 * 3 + "1\n" + sys.ps1


def test_overriden_compile():
    main(["tests/compile.py"])


@patch("time.strftime")
def test_bytecode_caching(patched_strftime):
    sys.dont_write_bytecode = False
    try:
        hello1 = "Hello, this file was cached at "
        hello2 = "Just hello."

        with open("tests/bytecode_test.pyh", "w") as f:
            f.write(f"#define HELLO {hello1!r} __TIME__")

        patched_strftime.return_value = "10:10:10"

        time_str = time.strftime("%H:%M:%S")
        hello1_full = hello1 + time_str + "\n"
        hello2_full = hello2 + "\n"

        with patch("sys.stdout", new=StringIO()):
            main(["tests/bytecode_test.ppy"])
            assert sys.stdout.getvalue() == hello1_full

        patched_strftime.return_value = "20:20:20"

        with patch("sys.stdout", new=StringIO()):
            main(["tests/bytecode_test.ppy"])
            assert sys.stdout.getvalue() == hello1_full

        time.sleep(0.01)

        with open("tests/bytecode_test.pyh", "w") as f:
            f.write(f"#define HELLO {hello2!r}")

        with patch("sys.stdout", new=StringIO()):
            main(["tests/bytecode_test.ppy"])
            assert sys.stdout.getvalue() == hello2_full
    finally:
        sys.dont_write_bytecode = True


def test_is_package():
    assert is_package("tests") is True
    assert is_package("tests.test_modules") is False
    with pytest.warns(match="Module file or directory not found"):
        assert is_package("inexistent") is False



import argparse
import os
import sys
from typing import Iterable
from importlib import util
from importlib.machinery import SourceFileLoader

from . import hooks
from .config import FILE_EXTENSIONS
from .version import __version__
from .utils import create_exception_handler, is_package


parser = argparse.ArgumentParser(
    (
        "python -m " + __package__
        if sys.argv[0] == "-m"
        else os.path.basename(sys.argv[0])
    ),
    description="Python with C preprocessor",
)
parser.add_argument(
    "--version", action="version", version="pwcp " + __version__
)
parser.add_argument("-m", action="store_true", help="run target as module")
parser.add_argument(
    "--prefer-py",
    dest="prefer_python",
    action="store_true",
    help="prefer .py files over .ppy when importing",
)
parser.add_argument(
    "--save-files",
    dest="save_files",
    action="store_true",
    help="save .ppy files to .py after preprocessing",
)
parser.add_argument(
    "--preprocess-unknown-sources",
    dest="preprocess_unknown_sources",
    action="store_true",
    help="preprocess code even if filename is unknown"
    " (for example, in exec call)",
)
parser.add_argument("target")
parser.add_argument("args", nargs=argparse.ZERO_OR_MORE)


def main_with_params(
    *,
    target: str,
    args: Iterable[str],
    m: bool,
    prefer_python: bool,
    save_files: bool,
    preprocess_unknown_sources: bool,
):
    hooks.install(
        prefer_python=prefer_python,
        save_files=save_files,
        preprocess_unknown_sources=preprocess_unknown_sources,
    )
    if not m:
        filename: str = os.path.abspath(target)
        sys.path.insert(0, os.path.dirname(filename))
        if filename.endswith(tuple(FILE_EXTENSIONS)):
            loader = hooks.PPyLoader
        else:
            loader = SourceFileLoader
        spec = util.spec_from_loader(
            "__main__",
            loader("__main__", filename),
        )
        vars_override = {"__package__": None}
    else:
        sys.path.insert(0, os.getcwd())
        if is_package(target):
            target += ".__main__"
        spec = util.find_spec(target)
        if spec is None:
            print("No module named " + target)
            return
        spec.loader.name = "__main__"
        vars_override = {"__name__": "__main__"}
    module = util.module_from_spec(spec)
    vars(module).update(vars_override)
    sys.modules["__main__"] = module
    sys.excepthook = create_exception_handler(module)
    orig_argv = sys.argv.copy()
    sys.argv.clear()
    sys.argv.append(module.__file__)
    sys.argv.extend(args)

    spec.loader.exec_module(module)

    sys.argv.clear()
    sys.argv.extend(orig_argv)


def main(args=sys.argv[1:]):
    args = parser.parse_args(args)
    main_with_params(**vars(args))


if __name__ == "__main__":
    main()



from io import StringIO
from linecache import getline
from importlib.machinery import SOURCE_SUFFIXES
from typing import Any, Callable, Optional, TextIO, Tuple, Union

from pypp import Preprocessor

from .config import FILE_EXTENSIONS
from .utils import py_from_ppy_filename
from .errors import PreprocessorError


preprocessed_files = {}


class PyPreprocessor(Preprocessor):
    default_disabled = True

    def __init__(self, disabled: Optional[bool] = None):
        if disabled is None:
            disabled = self.default_disabled
        super().__init__(disabled=disabled)
        self.included_files = []

    def write(self, file: TextIO):
        macros_backup = self.macros.copy()
        try:
            super().write(file)
        except Exception:
            self.macros = macros_backup
            raise

    def on_error(self, file: str, line: int, msg: str):
        raise SyntaxError(msg, (file, line, 1, getline(file, line)))

    def on_file_open(
        self, is_system_include: bool, includepath: str
    ) -> TextIO:
        self.included_files.append(includepath)
        return super().on_file_open(is_system_include, includepath)


PreprocessingFunction = Callable[[str, str, PyPreprocessor], str]


def _preprocess(src: str, filename: str, preprocessor: PyPreprocessor) -> str:
    preprocessor.parse(src, filename)

    out = StringIO()
    try:
        preprocessor.write(out)
    except SyntaxError:
        raise
    except Exception as e:
        msg = "internal preprocessor error"
        last = preprocessor.lastdirective
        if last:
            msg += f" at around {last.source}:{last.lineno}"
        raise PreprocessorError(msg) from e

    if preprocessor.return_code != 0:
        raise PreprocessorError(
            f"preprocessor exit code is not zero: {preprocessor.return_code}"
        )

    return out.getvalue()


def set_preprocessing_function(
    func: PreprocessingFunction,
) -> PreprocessingFunction:
    global _preprocess

    prev_func = _preprocess
    _preprocess = func

    return prev_func


def preprocess(
    src: Union[str, TextIO], filename: str, p: Optional[PyPreprocessor] = None
):
    if not isinstance(src, str):
        src = src.read()

    if p is None:
        # always enable preprocessing of ppy files
        if filename.endswith(tuple(FILE_EXTENSIONS)):
            disabled = False
        # but disable other Python files
        elif filename.endswith(tuple(SOURCE_SUFFIXES)):
            disabled = True
        else:
            disabled = None
        p = PyPreprocessor(disabled=disabled)

    # indicate that we started preprocessing
    preprocessed_files[filename] = None

    # save preprocessed file to display actual SyntaxError
    result = preprocessed_files[filename] = _preprocess(src, filename, p)
    return result, p.included_files


def preprocess_file(
    filename: str, save_files: bool = False
) -> Tuple[str, list]:
    with open(filename) as f:
        res, deps = preprocess(f, filename)
    if save_files:
        with open(py_from_ppy_filename(filename), "w") as f:
            f.write(res)
    return res, deps


def maybe_preprocess(
    src: Any, filename: str, preprocessor: Optional[PyPreprocessor] = None
):
    if isinstance(src, bytes):
        src = src.decode()
    if isinstance(src, str):
        # this is essential for interactive mode
        has_newline = src.endswith("\n")
        src, _ = preprocess(src, filename, preprocessor)
        if not has_newline:
            src = src.rstrip("\n")
    return src



__all__ = (
    "main",
    "main_with_params",
    "__version__",
    "add_file_extension",
    "install",
    "set_preprocessing_function",
)

from .runner import main, main_with_params
from .version import __version__
from .config import add_file_extension
from .hooks import install
from .preprocessor import set_preprocessing_function



import sys

import setuptools
import ppsetuptools
from setuptools import setup

sys.path.append("pwcp")
from version import __version__  # noqa: E402

del sys.path[-1]


def patched_setup(*args, **kwargs):
    scripts = kwargs.pop("scripts")
    kwargs["entry_points"] = {
        "console_scripts": [f"{k}={v}" for k, v in scripts.items()]
    }
    setup(*args, **{k: v for k, v in kwargs.items() if v is not None})


setuptools.setup = patched_setup
ppsetuptools.setup(
    version=__version__,
    packages=["pwcp"],
)
